from werkzeug.utils import import_string as wz_import_string, cached_property
from flask import abort
from slugify import slugify
import functools
import re
import yaml
import os
import subprocess
import click
import inspect


def import_string(impstr, attr=None):
    """Imports a string. Can import an attribute of the imported
    class/module using a double colon as a separator
    """
    if inspect.isclass(impstr):
        return impstr
    if "::" in impstr:
        impstr, attr = impstr.split("::")
    imported = wz_import_string(impstr)
    if attr is not None:
        return getobjpath(imported, attr)
    return imported


def getobjpath(obj, path):
    """Returns an item or attribute of the object recursively.
    Item names are specified between brackets, eg: [item].
    Attribute names are prefixed with a dot (the first one is optional), eg: .attr
    Example: getobjpath(obj, "attr1.attr2[item].attr3")
    """
    if not path:
        return obj
    if path.startswith("["):
        item = path[1:path.index("]")]
        return getobjpath(obj[item], path[len(item) + 2:])
    if path.startswith("."):
        path = path[1:]
    if "." in path or "[" in path:
        dot_idx = path.find(".")
        bracket_idx = path.find("[")
        if dot_idx == -1 or bracket_idx < dot_idx:
            idx = bracket_idx
            next_idx = idx
        else:
            idx = dot_idx
            next_idx = idx + 1
        attr = path[:idx]
        return getobjpath(getattr(obj, attr), path[next_idx:])
    return getattr(obj, path)


def find_classes_in_module(module, clstypes):
    """Find classes of clstypes in module
    """
    classes = []
    for item in dir(module):
        item = getattr(module, item)
        try:
            for cls in clstypes:
                if issubclass(item, cls) and item != cls:
                    classes.append(item)
        except:
            pass
    return classes


class ImportClassError(ImportError):
    pass


def find_class_in_module(cls, clstypes):
    if not isinstance(clstypes, (tuple, list)):
        clstypes = (clstypes,)

    if inspect.ismodule(cls):
        clsname = getattr(cls, '__frasco_extension__', None)
        if clsname:
            return getattr(cls, clsname)
        classes = find_classes_in_module(cls, clstypes)
        if not classes:
            raise ImportClassError('No extension class found in module')
        if len(classes) > 1:
            raise ImportClassError('Too many extension classes in module')
        return classes[0]

    if not issubclass(cls, clstypes):
        raise ImportClassError("Wrong class type")
    return cls


def import_class(impstr, clstypes, fallback_package=None):
    try:
        return find_class_in_module(import_string(impstr), clstypes)
    except ImportError as e:
        if not fallback_package or ('No module named' not in str(e) and str(e) != 'No extension class found in module'):
            raise
        return find_class_in_module(import_string(fallback_package + "." + impstr), clstypes)


def remove_yaml_frontmatter(source, return_frontmatter=False):
    """If there's one, remove the YAML front-matter from the source
    """
    if source.startswith("---\n"):
        frontmatter_end = source.find("\n---\n", 4)
        if frontmatter_end == -1:
            frontmatter = source
            source = ""
        else:
            frontmatter = source[0:frontmatter_end]
            source = source[frontmatter_end + 5:]
        if return_frontmatter:
            return (source, frontmatter)
        return source
    if return_frontmatter:
        return (source, None)
    return source


def parse_yaml_frontmatter(source):
    source, frontmatter = remove_yaml_frontmatter(source, True)
    if frontmatter:
        return (yaml.safe_load(frontmatter), source)
    return (None, source)


def populate_obj(obj, attrs):
    """Populates an object's attributes using the provided dict
    """
    for k, v in attrs.items():
        setattr(obj, k, v)


def make_kwarg_validator(name, validator_func):
    if not isinstance(name, tuple):
        name = (name,)
    def decorator_gen(**kwargs):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kw):
                values = {n: kw.get(n) for n in name}
                if not validator_func(**dict(kwargs, **values)):
                    abort(400)
                return func(*args, **kw)
            return wrapper
        return decorator
    return decorator_gen


def kwarg_validator(name):
    def decorator(validator_func):
        return make_kwarg_validator(name, validator_func)
    return decorator


def extract_unmatched_items(items, allowed_keys, prefix=None, uppercase=False):
    unmatched = {}
    for k, v in items.items():
        if k not in allowed_keys:
            if prefix:
                k = "%s%s" % (prefix, k)
            if uppercase:
                k = k.upper()
            unmatched[k] = v
    return unmatched


def deep_update_dict(a, b):
    for k, v in b.items():
        if k not in a:
            a[k] = v
        elif isinstance(a[k], dict) and isinstance(v, dict):
            deep_update_dict(a[k], v)
        elif isinstance(a[k], list) and isinstance(v, list):
            a[k].extend(v)
        elif isinstance(v, list) and not isinstance(a[k], list):
            a[k] = [a[k]] + v
        else:
            a[k] = v
    return a


class UnknownValue(object):
    pass

unknown_value = UnknownValue()


class AttrDict(dict):
    """Dict which keys are accessible as attributes
    """
    def __getattr__(self, name):
        if name not in self:
            raise AttributeError()
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]

    def get_or_raise(self, name, message):
        try:
            return self[name]
        except KeyError:
            raise KeyError(message)


class ShellError(Exception):
    def __init__(self, returncode, stderr):
        super(ShellError, self).__init__(stderr)
        self.returncode = returncode
        self.stderr = stderr

    def __str__(self):
        return str(self.stderr)


def shell_exec(args, echo=True, fg="green", **kwargs):
    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.STDOUT
    p = subprocess.run(args, **kwargs)
    if p.returncode > 0:
        raise ShellError(p.returncode, p.stdout)
    if echo:
        click.secho(p.stdout.decode(), fg=fg)
    return p.stdout


def match_domains(domain, allowed_domains):
    for allowed_domain in allowed_domains:
        if allowed_domain.startswith('^'):
            if re.search(allowed_domain, domain, re.I):
                return True
        elif allowed_domain.lower() == domain.lower():
            return True
    return False


def match_email_domain(email, allowed_domains):
    if not email or not '@' in email:
        return False
    _, domain = email.split('@')
    return match_domains(domain, allowed_domains)


def join_url_rule(rule1, rule2):
    if not rule1:
        return rule2
    if not rule2:
        return rule1
    return (rule1.rstrip('/') + '/' + rule2.lstrip('/')).rstrip('/')



from apispec import APISpec
from frasco.utils import join_url_rule, AttrDict
from frasco.request_params import RequestDataParam, get_marshmallow_schema, get_marshmallow_schema_instance
import re
import inspect


try:
    from marshmallow import Schema as MarshmallowSchema
    from marshmallow.fields import Field as MarshmallowField
    from marshmallow.exceptions import ValidationError as MarshmallowValidationError
    from apispec.ext.marshmallow import MarshmallowPlugin, resolver as ma_schema_name_resolver
    from apispec.ext.marshmallow.field_converter import DEFAULT_FIELD_MAPPING
    marshmallow_available = True
except ImportError:
    marshmallow_available = False


def build_openapi_spec(api_version, with_security_scheme=False):
    plugins = []
    if marshmallow_available:
        plugins.append(MarshmallowPlugin())

    spec = APISpec(title="API %s" % api_version.version,
                    version=api_version.version,
                    openapi_version="3.0.2",
                    plugins=plugins)

    if with_security_scheme:
        spec.components.security_scheme("api_key_bearer_token", {"type": "http", "scheme": "bearer"})
        spec.components.security_scheme("api_key_header", {"type": "apiKey", "in": "header", "name": "X-Api-Key"})
        spec.components.response("InvalidUserInput", {"description": "InvalidUserInput"})
        spec.components.response("NotAuthenticatedError", {"description": "Authentification required"})
        spec.components.response("NotAuthorizedError", {"description": "Some permissions are missing to perform this request"})

    for service in api_version.iter_services():
        paths = {}
        tag = {"name": service.name}
        if service.description:
            tag["description"] = service.description
        spec.tag(tag)
        for rule, endpoint, func, options in service.iter_endpoints():
            url = convert_url_args(join_url_rule(api_version.url_prefix, join_url_rule(service.url_prefix, rule)))
            path = paths.setdefault(url, {})
            for method in options.get('methods', ['GET']):
                op = build_spec_operation(spec, url, method, func, with_security_scheme)
                op.update({
                    'operationId': service.name + '_' + endpoint,
                    'tags': [service.name]
                })
                path[method.lower()] = op
        for path, operations in paths.items():
            spec.path(path=path, operations=operations)

    return spec


def build_spec_operation(spec, url, method, func, with_security_scheme=False):
    func_apispec = getattr(func, '__apispec__', {})

    for schema in func_apispec.get('schemas', []):
        get_schema_ref(spec, schema) # register schemas

    o = {}
    if 'description' in func_apispec:
        o['description'] = func_apispec['description']
    elif func.__doc__:
        o['description'] = func.__doc__

    if 'parameters' in func_apispec:
        o["parameters"] = func_apispec['parameters']
    elif hasattr(func, 'request_params'):
        query_params, body_params, file_params = build_spec_parameters(url, method, func.request_params)
        o["parameters"] = query_params
        request_body = build_spec_request_body_from_params(body_params, file_params)
        if request_body:
            o["requestBody"] = request_body

    if "requestBody" in func_apispec:
        o["requestBody"] = func_apispec["requestBody"]

    if "responses" in func_apispec:
        o["responses"] = func_apispec["responses"]
    else:
        o["responses"] = build_spec_responses(spec, func, with_security_scheme=with_security_scheme)

    return o


def build_spec_parameters(url, method, request_params):
    path_params = []
    search_params = []
    body_params = []
    file_params = []

    for p in reversed(request_params):
        if isinstance(p, RequestDataParam):
            if not marshmallow_available or not get_marshmallow_schema(p.loader):
                continue
            schema = get_marshmallow_schema(p.loader)
            only = None
            exclude = None
            if isinstance(p.loader, MarshmallowSchema):
                only = p.loader.only
                exclude = p.loader.exclude
            for name, field in schema._declared_fields.items():
                if only is not None and name not in only:
                    continue
                if exclude is not None and name in exclude:
                    continue
                if not field.dump_only:
                    body_params.append((name, field.required, convert_type_to_spec(field)))
            continue
        for pname in p.names:
            if p.location == 'files':
                file_params.append(pname)
            elif ("{%s}" % pname) in url:
                path_params.append(build_spec_param(pname, p, 'path'))
            elif method == 'GET':
                search_params.append(build_spec_param(pname, p))
            else:
                body_params.append((pname, p.required, convert_type_to_spec(p.type)))

    query_params = path_params + search_params
    return query_params, body_params, file_params


def build_spec_param(name, request_param, loc="query"):
    o = {"name": name,
        "schema": {"type": convert_type_to_spec(request_param.type)},
        "required": loc == "path" or bool(request_param.required),
        "in": loc}
    if request_param.help:
        o['description'] = request_param.help
    return o


def build_spec_request_body_from_params(body_params, file_params):
    request_body = {}
    if body_params:
        request_body["application/json"] = {"schema": {
            "type": "object",
            "required": [n for (n, r, t) in body_params if r],
            "properties": {n: {"type": t} for (n, r, t) in body_params}
        }}

    if file_params:
        request_body["multipart/form-data"] = {"schema": {
            "type": "object",
            "properties": {
                name: {"type": "string", "format": "binary"} \
                for name in file_params
            }
        }}

    if request_body:
        return {"content": request_body}


def build_spec_responses(spec, func, with_security_scheme=False):
    responses = {
        "default": {"description": "Unexpected error"},
        "200": {"description": "Successful response"},
        "400": {"$ref": "#/components/responses/InvalidUserInput"}
    }

    if hasattr(func, '__apispec__') and 'response' in func.__apispec__:
        responses["200"] = func.__apispec__['response']
    elif hasattr(func, '__apispec__') and 'response_content' in func.__apispec__:
        responses["200"]["content"] = func.__apispec__['response_content']
    elif hasattr(func, '__apispec__') and 'response_json' in func.__apispec__:
        responses["200"]["content"] = {"application/json": func.__apispec__['response_json']}
    else:
        content = build_spec_response_content(spec, func)
        if content:
            responses["200"]["content"] = content

    if with_security_scheme:
        responses["401"] = {"$ref": "#/components/responses/NotAuthenticatedError"}
        responses["403"] = {"$ref": "#/components/responses/NotAuthorizedError"}

    return responses


def build_spec_response_content(spec, func):
    if hasattr(func, 'marshalled_with'):
        response_spec = convert_response_to_spec(spec, func.marshalled_with, func.marshal_many)
        if response_spec:
            return {"application/json": {"schema": response_spec}}


def convert_response_to_spec(spec, response_type, many=False):
    if callable(response_type) and hasattr(response_type, '__apispec__') and 'spec' in response_type.__apispec__:
        return response_type.__apispec__['spec']

    if marshmallow_available and get_marshmallow_schema(response_type):
        schema = get_marshmallow_schema(response_type)
        return build_spec_from_marshmallow_schema(spec, schema, many)

    if isinstance(response_type, dict):
        schema = {"type": "object", "properties": {}}
        for key, value in response_type.items():
            if marshmallow_available and get_marshmallow_schema(value):
                schema["properties"][key] = build_spec_from_marshmallow_schema(spec, get_marshmallow_schema_instance(value))
            elif callable(value):
                prop = convert_response_to_spec(spec, value)
                schema["properties"][key] = prop if prop else {"type": "string"}
            else:
                schema["properties"][key] = {"type": convert_type_to_spec(value)}
        return schema


def build_spec_from_marshmallow_schema(spec, schema, many=False):
    schema_ref = get_schema_ref(spec, schema.__class__ if not inspect.isclass(schema) else schema)
    if many or isinstance(schema, MarshmallowSchema) and schema.many:
        return {"type": "array", "items": schema_ref}
    return schema_ref


def get_schema_ref(spec, schema):
    schema_name = ma_schema_name_resolver(schema)
    if schema_name not in spec.components.schemas:
        spec.components.schema(schema_name, schema=schema)
    return {"$ref": "#/components/schemas/%s" % schema_name}


_url_arg_re = re.compile(r"<([a-z]+:)?([a-z0-9_]+)>")
def convert_url_args(url):
    return _url_arg_re.sub(r"{\2}", url)


def convert_type_to_spec(type):
    if marshmallow_available and isinstance(type, MarshmallowField):
        return DEFAULT_FIELD_MAPPING.get(type.__class__, ('string', None))[0]
    if type is int:
        return "integer"
    if type is float:
        return "number"
    if type is bool:
        return "boolean"
    return "string"


def api_spec(**components):
    def decorator(func):
        func.__apispec__ = components
        return func
    return decorator



from sqlalchemy import event
from sqlalchemy.orm.session import object_session
from .transactions import current_transaction
import datetime
import functools
import inspect
from flask import current_app


def listens_for_modified_object_before_update(cls, func=None, include_collections=False):
    def decorator(func):
        @event.listens_for(cls, "before_update")
        def before_update(mapper, connection, target):
            if object_session(target).is_modified(target, include_collections=include_collections):
                func(target)
    if func:
        decorator(func)
        return
    return decorator


def auto_updated_at(cls):
    @listens_for_modified_object_before_update(cls)
    def on_update(target):
        target.updated_at = datetime.datetime.utcnow()
    return cls


def call_once_per_transaction(func):
    @functools.wraps(func)
    def wrapper(self):
        already_called = current_transaction.setdefault('call_once', {}).setdefault(func, set())
        if self not in already_called:
            already_called.add(self)
            func(self)
    return wrapper


def always_call_once_per_transaction_when_update(funcname):
    def decorator(cls):
        setattr(cls, funcname, call_once_per_transaction(getattr(cls, funcname)))

        @listens_for_modified_object_before_update(cls)
        def before_update(target):
            getattr(target, funcname)()

        return cls
    return decorator


def auto_clear_cache(cls):
    return always_call_once_per_transaction_when_update('clear_cache')(cls)



from flask import g, current_app
from flask.signals import Namespace
from frasco.ext import *
from frasco.models import delayed_tx_calls
from frasco.tasks import enqueue_task
from frasco.templating.extensions import RemoveYamlFrontMatterExtension
from frasco.utils import extract_unmatched_items, import_class, AttrDict
from jinja_macro_tags import MacroLoader, MacroRegistry
from jinja2 import ChoiceLoader, FileSystemLoader, PackageLoader
from contextlib import contextmanager
from flask_mail import Mail
import os
import datetime

from .message import create_message, clickable_links, markdown_jinja, log_message
from .provider import MailProvider, bulk_connection_context


_signals = Namespace()
email_sent = _signals.signal('email_sent')


class FrascoMailError(Exception):
    pass


class FrascoMailState(ExtensionState):
    def __init__(self, *args):
        super(FrascoMailState, self).__init__(*args)
        self.connections = AttrDict()
        self.template_loaders = []
        layout_loader = PackageLoader(__name__, "templates")
        self.jinja_loader = MacroLoader(ChoiceLoader([ChoiceLoader(self.template_loaders), layout_loader]))
        self.jinja_env = None

    def add_template_folder(self, path):
        self.template_loaders.append(FileSystemLoader(path))

    def add_templates_from_package(self, pkg_name, pkg_path="emails"):
        self.template_loaders.append(PackageLoader(pkg_name, pkg_path))


class FrascoMail(Extension):
    """Send emails using SMTP
    """
    name = "frasco_mail"
    state_class = FrascoMailState
    defaults = {"provider": "smtp",
                "connections": {},
                "default_layout": "layout.html",
                "default_template_vars": {},
                "inline_css": False,
                "auto_render_missing_content_type": True,
                "log_messages": None, # default is app.testing
                "log_messages_on_failure": False, # default is app.testing
                "dumped_messages_folder": "email_logs",
                "localized_emails": None,
                "default_locale": None,
                "markdown_options": {},
                "suppress_send": False,
                "silent_failures": False,
                "send_async": False}

    def _init_app(self, app, state):
        mail = Mail(app) # needed for Message even if we will never use
        connections = dict(state.options['connections'])
        if "default" not in connections:
            connections["default"] = dict(extract_unmatched_items(state.options, self.defaults or {}),
                provider=state.require_option('provider'))
        for name, opts in connections.items():
            self.register_connection(name, opts, _app=app)

        state.add_template_folder(os.path.join(app.root_path, "emails"))
        state.jinja_env = app.jinja_env.overlay(loader=state.jinja_loader)
        state.jinja_env.add_extension(RemoveYamlFrontMatterExtension)
        state.jinja_env.macros = MacroRegistry(state.jinja_env) # the overlay methods does not call the constructor of extensions
        state.jinja_env.macros.register_from_template("layouts/macros.html")
        state.jinja_env.default_layout = state.options["default_layout"]
        state.jinja_env.filters['clickable_links'] = clickable_links
        state.jinja_env.filters['markdown'] = markdown_jinja

        if has_extension('frasco_babel', app):
            if state.options['default_locale'] is None:
                state.options['default_locale'] = app.extensions.frasco_babel.options['default_locale']
            if state.options['localized_emails'] is None:
                state.options['localized_emails'] = '{locale}/{filename}'

    @ext_stateful_method
    def create_connection(self, state, options, provider=None):
        if not provider:
            options = dict(**options)
            provider = options.pop('provider', 'smtp')
        provider_class = import_class(provider, MailProvider, "frasco.mail.providers")
        return provider_class(state, options)

    @ext_stateful_method
    def register_connection(self, state, name, options, provider=None):
        state.connections[name] = self.create_connection(options, provider, _state=state)


@delayed_tx_calls.proxy
def send_message_sync(msg, connection="default", silent=None):
    state = get_extension_state('frasco_mail')

    if state.options['suppress_send']:
        if state.options["log_messages"] or current_app.testing or current_app.debug:
            log_message(msg, state.options['dumped_messages_folder'], connection)
        email_sent.send(msg)
        return

    try:
        if bulk_connection_context.top:
            bulk_connection_context.top.send(msg)
        elif isinstance(connection, MailProvider):
            connection.send(msg)
        else:
            state.connections[connection].send(msg)
        email_sent.send(msg)
        if state.options["log_messages"] or current_app.testing or current_app.debug:
            log_message(msg, state.options['dumped_messages_folder'], connection)
    except Exception as e:
        if state.options["log_messages_on_failure"]:
            log_message(msg, state.options['dumped_messages_folder'], connection)
        if (silent is None and not state.options['silent_failures']) or silent is False:
            raise e
        current_app.log_exception(e)


@delayed_tx_calls.proxy
def send_message_async(msg, connection="default"):
    require_extension('frasco_tasks')
    enqueue_task(send_message_sync, msg=msg, connection=connection)


def send_message(msg, connection="default"):
    state = get_extension_state('frasco_mail')
    if has_extension('frasco_tasks') and state.options['send_async']:
        send_message_async(msg, connection)
    else:
        send_message_sync(msg, connection)


def send_mail(to, template_filename, *args, **kwargs):
    state = get_extension_state('frasco_mail')
    force_sync = kwargs.pop('_force_sync', False)
    connection = kwargs.pop('_connection', 'default')
    msg = create_message(to, template_filename, *args, **kwargs)
    if msg:
        if force_sync:
            send_message_sync(msg, connection)
        else:
            send_message(msg, connection)

@contextmanager
def bulk_mail_connection(connection="default"):
    state = get_extension_state('frasco_mail')
    with state.connections[connection].bulk_connection():
        yield



from frasco.ext import get_extension_state
from frasco.models import db
from ..password import check_password


AUTH_HANDLERS = []


def register_authentification_handler(func=None, only=False):
    def decorator(f):
        if only:
            AUTH_HANDLERS[:] = []
        AUTH_HANDLERS.append(f)
        return f
    if func:
        return decorator(func)
    return decorator


def authenticate(identifier, password):
    state = get_extension_state('frasco_users')
    for func in AUTH_HANDLERS:
        user = func(identifier, password)
        if user:
            return user

    if not state.options["disable_password_authentication"]:
        if state.options['allow_email_or_username_login'] and hasattr(state.Model, 'username'):
            q = state.Model.query.filter(db.or_(db.func.lower(state.Model.username) == identifier.strip().lower(),
                state.Model.email == identifier.strip().lower()))
        else:
            q = state.Model.query_by_identifier(identifier)
        user = q.first()
        if user and check_password(user, password):
            return user




from flask import current_app, has_app_context
from frasco.utils import (AttrDict, import_string, find_classes_in_module, extract_unmatched_items,
                          import_string, import_class, deep_update_dict)
from collections import OrderedDict
import functools
import inspect
import logging


__all__ = ('Extension', 'ExtensionState', 'ExtensionError', 'ext_stateful_method', 'pass_extension_state',
           'require_extension', 'has_extension', 'get_extension_state')


class ExtensionError(Exception):
    pass


class ExtensionState(object):
    def __init__(self, extension, options):
        self.extension = extension
        self.options = AttrDict(options)

    def get_option(self, name, default=None):
        return self.options.get(name, default)

    def require_option(self, name):
        if not self.options.get(name):
            raise ExtensionError('Option "%s" is missing for extension "%s"' % (name, self.extension.name))
        return self.options[name]

    def import_option(self, name, required=True):
        if required:
            import_name = self.require_option(name)
        else:
            import_name = self.get_option(name)
        if import_name:
            return import_string(import_name)

    def import_option_as_class(self, name, clstypes, fallback_package=None, required=True):
        if required:
            import_name = self.require_option(name)
        else:
            import_name = self.get_option(name)
        if import_name:
            return import_class(import_name, clstypes, fallback_package)


class Extension(object):
    name = None
    defaults = None
    prefix_extra_options = None
    state_class = ExtensionState

    def __init__(self, app=None, **options):
        if not self.name:
            raise ExtensionError("A name property must be defined on Extension classes")
        self._wait_for_state_methods = []
        self.app = app
        self.init_options = options
        if app:
            self.init_app(app)

    def init_app(self, app, **options):
        options = deep_update_dict(deep_update_dict(dict(self.init_options), options), get_extension_config(app, self.name))
        if self.prefix_extra_options:
            app.config.update(extract_unmatched_items(options, self.defaults or {},
                prefix=self.prefix_extra_options, uppercase=True))
        app.extensions[self.name] = self.state_class(self, dict(self.defaults or {}, **options))
        self._init_app(app, app.extensions[self.name])
        for func, args, kwargs in self._wait_for_state_methods:
            func(self, app.extensions[self.name], *args, **kwargs)

    def _init_app(self, app, state):
        raise NotImplementedError()

    def get_app(self, app=None):
        if app is not None:
            return app
        if current_app:
            return current_app._get_current_object()
        if self.app is not None:
            return self.app
        raise RuntimeError(
            'No application found. Either work inside a view function or push'
            ' an application context.'
        )

    def get_state(self, app=None):
        return get_extension_state(self.name, app=self.get_app(app))


def ext_stateful_method(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        state = kwargs.pop('_state', None)
        if not state:
            app = kwargs.pop('_app', None)
            if has_app_context() or self.app or app:
                state = self.get_state(app)
        if state:
            return func(self, state, *args, **kwargs)
        self._wait_for_state_methods.append((func, args, kwargs))
    return wrapper


def get_extension_config(app, name):
    opts = {k.lower(): v for k, v in app.config.get(name.upper(), {}).items()}
    opts.update(app.config.get_namespace("%s_" % name.upper()))
    return opts


def get_extension_state(ext_name, state=None, app=None, must_exist=True):
    if state:
        return state
    if not app:
        app = current_app
    if must_exist:
        assert ext_name in app.extensions, ("The current app does not have the extension '%s'" % ext_name)
    return app.extensions.get(ext_name)


def pass_extension_state(ext_name=None):
    def decorator(func):
        spec = inspect.getargspec(func)
        if spec.args and spec.args[0] == 'self':
            def wrapper(self, *args, **kwargs):
                app = kwargs.pop('_app', None)
                if isinstance(self, Extension):
                    return func(self, self.get_state(app), *args, **kwargs)
                return func(self, get_extension_state(ext_name, app=app), *args, **kwargs)
        else:
            def wrapper(*args, **kwargs):
                app = kwargs.pop('_app', None)
                return func(get_extension_state(ext_name, app=app), *args, **kwargs)
        return functools.wraps(func)(wrapper)
    return decorator


def require_extension(ext_name, app=None):
    get_extension_state(ext_name, app)


def has_extension(name, app=None):
    return get_extension_state(name, app=app, must_exist=False) is not None


def load_extensions_from_config(app, key='EXTENSIONS', **kwargs):
    load_extensions(app, app.config.get(key, []), **kwargs)


def load_extensions(app, config, aliases=None, fallback_package=None):
    loaded = []
    exts = list_extensions_from_config(app, config, aliases)
    for ext_module, options in exts.items():
        ext_class = import_class(ext_module, Extension, fallback_package)
        loaded.append(ext_class(app, **options))
        logging.getLogger('frasco').info("Extension '%s.%s' loaded" % (ext_class.__module__, ext_class.__name__))
    return loaded


def list_extensions_from_config(app, config, aliases=None):
    if not aliases:
        aliases = {}
    exts = OrderedDict()
    for spec in config:
        if isinstance(spec, str):
            ext_module = spec
            options = {}
        elif isinstance(spec, (tuple, list)):
            ext_module, options = spec
        elif isinstance(spec, dict):
            ext_module, options = next(x for x in spec.items())
        exts[aliases.get(ext_module, ext_module)] = options
    return exts



from flask import *
from .app import Frasco
from .config import load_config
from .request_params import (request_param, request_params, partial_request_param,
                             partial_request_param_loader, disable_request_params)
from .marshaller import marshal_with, marshal_many_with, marshal_dict_with, disable_marshaller, marshalling_context
from .ext import has_extension, require_extension, load_extensions_from_config
from .i18n import translate, translatable, lazy_translate, ntranslate, format_date, format_datetime, format_time
from .helpers import url_for, url_for_same, url_for_static, get_remote_addr, wrap_in_markup, inject_app_config



from flask import current_app, abort
from frasco.utils import unknown_value
from sqlalchemy.ext.mutable import Mutable
from flask_sqlalchemy import Pagination
import uuid
import functools
import sqlalchemy
from .ext import db


def move_obj_position_in_collection(obj, new_position, position_attr='position', scope=None, data=None, current_position=unknown_value):
    if not data:
        data = {}
    if current_position is unknown_value:
        current_position = getattr(obj, position_attr, None)
    shift, lower_idx, upper_idx = compute_move_obj_position_in_collection_bounds(current_position, new_position)
    position_col = getattr(obj.__class__, position_attr)

    q = obj.__class__.query
    if scope:
        q = q.filter_by(**scope)
    q = q.filter(position_col >= lower_idx, position_col <= upper_idx)
    q.update(dict(dict([(position_col.name, position_col + shift)]), **data))
    setattr(obj, position_attr, new_position)


def compute_move_obj_position_in_collection_bounds(current_position, new_position):
    if current_position is None:
        return 1, new_position, None
    up = new_position > current_position
    shift = -1 if up else 1
    lower_idx = min(current_position, new_position)
    lower_idx += 1 if up else 0
    upper_idx = max(current_position, new_position)
    upper_idx -= 0 if up else 1
    return shift, lower_idx, upper_idx


def ensure_unique_value(model, column, value, fallback=None, counter_start=1):
    if not fallback:
        fallback = value + "-%(counter)s"
    counter = counter_start
    q = model.query
    while q.filter_by(**dict([(column, value)])).count() > 0:
        value = fallback % {'counter': counter}
        counter += 1
    return value


def clone_sqla_obj(obj, propmapping=None):
    if not propmapping:
        propmapping = {}
    new = obj.__class__()
    for c in obj.__table__.columns:
        prop = propmapping.get(c.name, c.name)
        if prop == 'cache_id':
            new.cache_id = uuid.uuid4()
        elif prop != 'id' and hasattr(obj, prop):
            setattr(new, prop, getattr(obj, prop))
    return new


class MutableList(Mutable, list):
    def append(self, value):
        list.append(self, value)
        self.changed()

    def remove(self, value):
        list.remove(self, value)
        self.changed()

    @classmethod
    def coerce(cls, key, value):
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)
            return Mutable.coerce(key, value)
        else:
            return value


def ensure_session_flushed(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        db.session.flush()
        return func(*args, **kwargs)
    return wrapper


def model_loader(model, nullable=False, *options, **options_kwargs):
    """Model loader of request params"""
    def loader(id):
        if id is None:
            if not nullable:
                abort(404)
            return None
        q = model.query
        for opt in options:
            q = q.options(opt)
        for item in options_kwargs.pop('joinedload', []):
            q = q.options(db.joinedload(item))
        for item in options_kwargs.pop('undefer', []):
            q = q.options(db.undefer_group(item))
        for method, value in options_kwargs.items():
            q = q.options(getattr(db, method)(value))
        return q.get_or_404(id)
    return loader


def extract_model_attr_to_col_mapping(model):
    mapper = sqlalchemy.inspect(model)
    mapping = {}
    for attr in mapper.column_attrs:
        if not isinstance(attr.expression, sqlalchemy.sql.schema.Column):
            continue
        mapping[attr.key] = attr.expression.name
    return mapping


def model_obj_to_dict(obj, value_serializer=None):
    row = {}
    for attr, col in extract_model_attr_to_col_mapping(obj.__class__).items():
        row[col] = (value_serializer or (lambda a: a))(getattr(obj, attr))
    return row


def get_model_class(name):
    if name in db.Model._decl_class_registry:
        return db.Model._decl_class_registry[name]


def get_model_class_from_table(name):
    for model in db.Model._decl_class_registry:
        if model.__table__.name == name:
            return model



from flask import Blueprint, request, redirect, current_app, flash, session, render_template, abort, make_response, jsonify, get_flashed_messages
from flask_login import logout_user, current_user
from frasco.helpers import url_for
from frasco.ext import get_extension_state, has_extension
from frasco.utils import populate_obj, AttrDict
from frasco.models import db, transaction
from frasco.mail import send_mail
from frasco.geoip import geolocate_country
import datetime
import requests
import logging
import re

from .forms import *
from .auth import authenticate
from .auth.oauth import clear_oauth_signup_session
from .signals import user_signed_up, email_validated
from .user import is_user_logged_in, login_user, signup_user, UserValidationFailedError, check_rate_limit, send_user_validation_email, validate_user_email, send_user_welcome_email
from .password import generate_reset_password_token, update_password, validate_password, send_reset_password_token, PasswordValidationFailedError
from .tokens import read_user_token, generate_user_token, TOKEN_NS_ACCESS_TOKEN, TOKEN_NS_2FA, TOKEN_NS_PASSWORD_RESET, TOKEN_NS_VALIDATE_EMAIL
from .otp import verify_2fa


users_blueprint = Blueprint("users", __name__, template_folder="templates")
logger = logging.getLogger('frasco.users')


LOGIN_FLOW_ACCESS_TOKEN = 'access_token'
LOGIN_FLOW_WEB_ACCESS_TOKEN = 'web_access_token'


def _is_login_flow(flow):
    return request.args.get('flow') == flow


def _is_programmatic_login():
    return request.is_json or _is_login_flow(LOGIN_FLOW_ACCESS_TOKEN)


def _login_redirect(url, **kwargs):
    if request.method == 'POST' and _is_programmatic_login():
        kwargs.setdefault('errors', get_flashed_messages())
        return jsonify(success=False, next=url, **kwargs)
    return redirect(url)


def _do_login(user, remember=None):
    with transaction():
        if "oauth_signup" in session:
            user.save_oauth_token_data(session['oauth_signup'], session['oauth_data'])
        login_user(user, remember=remember, provider=session.get('oauth_signup'), skip_session=_is_programmatic_login() or _is_login_flow(LOGIN_FLOW_WEB_ACCESS_TOKEN))
        clear_oauth_signup_session()


def _login_success(opt_redirect='redirect_after_login', allow_programmatic=True):
    state = get_extension_state('frasco_users')
    redirect_url = request.args.get("next") or _make_redirect_url(state.options[opt_redirect])
    if _is_login_flow(LOGIN_FLOW_WEB_ACCESS_TOKEN) and request.args.get("next") and state.options['enable_access_tokens'] and state.options['enable_access_tokens_web_flow']:
        for allowed_url_pattern in state.options['access_tokens_web_flow_allowed_redirects']:
            if re.match(allowed_url_pattern, redirect_url, re.I):
                redirect_url += '#access_token=%s' % generate_user_token(current_user, TOKEN_NS_ACCESS_TOKEN)
                break
    if allow_programmatic and _is_programmatic_login():
        return jsonify(success=True)
    return redirect(redirect_url)


@users_blueprint.route('/login', methods=['GET', 'POST'])
def login():
    state = get_extension_state('frasco_users')

    if request.method == "GET" and not "oauth" in request.args:
        # signup was accessed directly so we ensure that oauth
        # params stored in session are cleaned. this can happen
        # if a user started to login using an oauth provider but
        # didn't complete the process
        clear_oauth_signup_session()

    is_oauth = "oauth_signup" in session

    if is_user_logged_in():
        return _login_success()

    if not is_oauth and not request.args.get("no_redirect"):
        if state.options['login_redirect']:
            return _login_redirect(state.options['login_redirect'])
        if state.manager.login_view != "users.login":
            return _login_redirect(url_for(state.manager.login_view, next=request.args.get("next")))

    form = state.import_option('login_form_class')(meta={'csrf': current_app.config.get('WTF_CSRF_ENABLED', True) and not _is_programmatic_login()})

    if not state.options['allow_login']:
        if state.options["login_disallowed_message"]:
            flash(state.options["login_disallowed_message"], "error")
        if request.method == 'POST' and _is_programmatic_login():
            return jsonify(success=False, errors=get_flashed_messages())
        if state.options.get("redirect_after_login_disallowed"):
            return _login_redirect(_make_redirect_url(state.options.get("redirect_after_login_disallowed")))
    elif request.method == 'POST':
        user = None
        try:
            if request.is_json:
                payload = request.get_json()
                user = authenticate(payload.get('identifier'), payload.get('password'))
            elif form.validate():
                user = authenticate(form.identifier.data, form.password.data)
        except Exception as e:
            current_app.log_exception(e)

        if user:
            if (state.options['expire_password_after'] and user.last_password_change_at and \
              (datetime.datetime.utcnow() - user.last_password_change_at).total_seconds() > state.options['expire_password_after']) \
              or user.must_reset_password_at_login:
                token = generate_reset_password_token(user)
                flash(state.options['password_expired_message'], 'error')
                return _login_redirect(url_for('.reset_password', token=token))

            if state.options['enable_2fa'] and user.two_factor_auth_enabled:
                if _is_programmatic_login():
                    return jsonify(success=False, require_2fa=True)
                session['2fa'] = user.id
                session['login_remember'] = form.remember.data
                return redirect(url_for('.login_2fa', flow=request.args.get('flow'), next=request.args.get('next')))

            if state.options['block_non_email_validated_users'] and user.must_validate_email:
                clear_oauth_signup_session()
                return render_template("users/non_email_validated_users_block_page.html", email=user.email)

            _do_login(user, form.remember.data)

            if state.options['enable_access_tokens'] and _is_programmatic_login():
                return jsonify(success=True, access_token=generate_user_token(user, TOKEN_NS_ACCESS_TOKEN))
            
            return _login_success()

        flash(state.options['login_error_message'], 'error')

        if _is_programmatic_login():
            return jsonify(success=False, errors=get_flashed_messages())

    return render_template('users/login.html', form=form, is_oauth=is_oauth)


@users_blueprint.route('/login/2fa', methods=['GET', 'POST'])
def login_2fa():
    state = get_extension_state('frasco_users')
    if not state.options['enable_2fa'] or not session.get('2fa'):
        return redirect(url_for('.login', flow=request.args.get('flow'), next=request.args.get('next')))

    is_oauth = "oauth_signup" in session
    remember_2fa_max_age = state.options['2fa_remember_days']*3600*24

    if request.cookies.get('remember_2fa'):
        remember_user = read_user_token(request.cookies['remember_2fa'], TOKEN_NS_2FA, remember_2fa_max_age)
        user = state.Model.query.get(session.pop('2fa'))
        if remember_user == user:
            _do_login(user, session.pop('login_remember'))
            return _login_success()

    form = state.import_option('login_2fa_form_class')()
    if form.validate_on_submit():
        user = state.Model.query.get(session.pop('2fa'))
        remember = session.pop('login_remember')
        if verify_2fa(user, form.code.data):
            _do_login(user, remember)
            r = make_response(_login_success())
            if form.remember.data:
                r.set_cookie('remember_2fa', generate_user_token(user, TOKEN_NS_2FA),
                    max_age=remember_2fa_max_age, **state.options['2fa_remember_cookie_options'])
            return r
        flash(state.options['login_2fa_error_message'], 'error')
        return redirect(url_for('.login', flow=request.args.get('flow'), next=request.args.get('next')))

    return render_template('users/login_2fa.html', form=form, is_oauth=is_oauth)


@users_blueprint.route('/logout')
def logout():
    state = get_extension_state('frasco_users')
    redirect_url = request.args.get("next") or _make_redirect_url(state.options["redirect_after_logout"])
    logout_user()
    return redirect(redirect_url)


@users_blueprint.route('/signup', methods=['GET', 'POST'])
def signup():
    state = get_extension_state('frasco_users')
    if request.method == "GET" and not "oauth" in request.args:
        # signup was accessed directly so we ensure that oauth
        # params stored in session are cleaned. this can happen
        # if a user started to login using an oauth provider but
        # didn't complete the process
        clear_oauth_signup_session()

    if is_user_logged_in():
        return _login_success('redirect_after_signup', False)

    if state.options['signup_redirect']:
        return redirect(state.options['signup_redirect'])
        
    is_oauth = "oauth_signup" in session
    allow_signup = state.options["allow_signup"]
    if state.options["oauth_signup_only"] and not is_oauth:
        allow_signup = False

    if not allow_signup:
        if state.options["signup_disallowed_message"]:
            flash(state.options["signup_disallowed_message"], "error")
        return redirect(url_for(state.options.get("redirect_after_signup_disallowed") or\
            "users.login", next=request.args.get("next")))

    form = state.import_option('signup_form_class')(obj=AttrDict(session.get('oauth_user_defaults', {})))
    must_provide_password = not is_oauth or state.options["oauth_must_provide_password"]

    if form.validate_on_submit():
        try:
            if must_provide_password and not form.password.data:
                raise PasswordValidationFailedError()

            if not check_rate_limit(request.remote_addr, 'signup_from', 'signup_at'):
                raise UserValidationFailedError()

            if state.captcha_validator and (not current_app.debug and not current_app.testing or state.options['debug_captcha']):
                is_captcha_success = False
                try:
                    is_captcha_success = state.captcha_validator()
                except:
                    logger.exception("Error while checking recaptcha")
                if not is_captcha_success:
                    if state.options['captcha_fail_message']:
                        flash(state.options['captcha_fail_message'], 'error')
                    raise UserValidationFailedError()

            with transaction():
                user = state.Model()
                populate_obj(user, session.get("oauth_user_defaults", {}))
                signup_user(user, provider=session.get('oauth_signup'), send_signal=False, **form.data)
                if is_oauth:
                    _save_oauth(user)
                db.session.flush()
                user_signed_up.send(user=user)
                if state.options['block_non_email_validated_users'] and not user.email_validated:
                    clear_oauth_signup_session()
                    return render_template("users/non_email_validated_users_block_page.html", email=user.email)
                if state.options["login_user_on_signup"]:
                    login_user(user, provider=user.signup_provider, skip_session=_is_programmatic_login() or _is_login_flow(LOGIN_FLOW_WEB_ACCESS_TOKEN))

            clear_oauth_signup_session()
            return _login_success('redirect_after_signup', False)
        except (UserValidationFailedError, PasswordValidationFailedError):
            db.session.rollback()

    return render_template('users/signup.html',
        form=form, must_provide_password=must_provide_password, is_oauth=is_oauth)


@users_blueprint.route('/signup/oauth')
def oauth_signup():
    state = get_extension_state('frasco_users')
    if "oauth_signup" not in session or state.options["oauth_must_signup"]:
        oauth = 1 if state.options["oauth_must_signup"] else 0
        return redirect(url_for(".signup", oauth=oauth, flow=request.args.get('flow'), next=request.args.get("next")))

    try:
        with transaction():
            user = state.Model()
            populate_obj(user, session.get("oauth_user_defaults", {}))
            signup_user(user, flash_messages=False, send_signal=False)
            _save_oauth(user)
            db.session.flush()
            user_signed_up.send(user=user)
            login_user(user, provider=session['oauth_signup'])
    except UserValidationFailedError:
        return redirect(url_for(".signup", oauth=1, flow=request.args.get('flow'), next=request.args.get("next")))

    clear_oauth_signup_session()
    return _login_success()


def _save_oauth(user):
    user.save_oauth_token_data(session['oauth_signup'], session['oauth_data'])
    if session.get("oauth_validate_email") and user.email == session.get("oauth_user_defaults", {}).get('email') and hasattr(user, 'email_validated'):
        validate_user_email(user)


@users_blueprint.route('/login/reset-password', methods=['GET', 'POST'])
def send_reset_password():
    state = get_extension_state('frasco_users')

    if state.options['reset_password_redirect']:
        return redirect(state.options['reset_password_redirect'])

    if not state.options['allow_reset_password']:
        if state.options["reset_password_disallowed_message"]:
            flash(state.options["reset_password_disallowed_message"], "error")
        return redirect(url_for(state.options.get("redirect_after_reset_password_disallowed") or\
            "users.login", next=request.args.get("next")))

    success_msg = state.options["reset_password_token_success_message"]
    error_msg = state.options["reset_password_token_error_message"]
    redirect_to = state.options["redirect_after_reset_password_token"]

    form = state.import_option('send_reset_password_form_class')()
    if form.validate_on_submit():
        user = state.Model.query_by_email(form.email.data).first()
        if user:
            send_reset_password_token(user)
            if success_msg:
                flash(success_msg, "success")
            if redirect_to:
                return redirect(_make_redirect_url(redirect_to))
        elif error_msg:
            flash(error_msg, 'error')

    return render_template("users/send_reset_password.html", form=form)


@users_blueprint.route('/login/reset-password/<token>', methods=['GET', 'POST'])
def reset_password(token):
    state = get_extension_state('frasco_users')

    if not state.options['allow_reset_password']:
        if state.options["reset_password_disallowed_message"]:
            flash(state.options["reset_password_disallowed_message"], "error")
        return redirect(url_for(state.options.get("redirect_after_reset_password_disallowed") or\
            "users.login", next=request.args.get("next")))

    msg = state.options["reset_password_success_message"]
    redirect_to = state.options["redirect_after_reset_password"]

    user = read_user_token(token, TOKEN_NS_PASSWORD_RESET, state.options['reset_password_ttl'])
    if not user:
        abort(404)

    form = state.import_option('reset_password_form_class')()
    if form.validate_on_submit() and validate_password(user, form.password.data, raise_error=False):
        with transaction():
            update_password(user, form.password.data, skip_validation=True)
            if state.options['login_user_on_reset_password']:
                login_user(user)
            send_mail(user.email, "users/reset_password_done", locale=getattr(user, 'locale', None))
            if msg:
                flash(msg, "success")
            if redirect_to:
                return redirect(_make_redirect_url(redirect_to))

    return render_template("users/reset_password.html", form=form)


@users_blueprint.route('/validate-email/send', methods=['POST'])
def send_email_validation_email():
    state = get_extension_state('frasco_users')
    user = state.Model.query.filter_by(email=request.form['email'].lower()).first_or_404()
    if user.email_validated:
        abort(403)

    send_user_validation_email(user)

    if request.values.get('next'):
        return redirect(request.values['next'])

    return render_template("users/non_email_validated_users_block_page.html", email=user.email)


@users_blueprint.route('/validate-email/<token>', methods=['GET'])
def validate_email(token):
    state = get_extension_state('frasco_users')
    msg = state.options["email_validation_success_message"]
    redirect_to = state.options["redirect_after_email_validated"]

    user = read_user_token(token, TOKEN_NS_VALIDATE_EMAIL, state.options['email_validation_ttl'])
    if not user:
        abort(404)

    send_welcome_email = state.options["send_welcome_email"] and state.options['block_non_email_validated_users'] and user.must_validate_email

    with transaction():
        validate_user_email(user)

    if send_welcome_email:
        send_user_welcome_email(user)

    if not current_user.is_authenticated and state.options["login_after_email_validation"]:
        login_user(user)

    if msg:
        flash(msg, "success")
    return redirect(_make_redirect_url(redirect_to))


def _make_redirect_url(value):
    if value.startswith('http://') or value.startswith('https://'):
        return value
    return url_for(value)



from flask import current_app, json
from frasco.ext import get_extension_state
from frasco.utils import shell_exec
from babel.messages import pofile
import os
import tempfile
import contextlib


@contextlib.contextmanager
def edit_pofile(filename, autosave=True):
    with open(filename, "r") as f:
        catalog = pofile.read_po(f)
    yield catalog
    if autosave:
        with open(filename, "wb") as f:
            pofile.write_po(f, catalog)


def merge_pofile(filename1, filename2):
    with edit_pofile(filename1) as catalog1:
        with edit_pofile(filename2) as catalog2:
            for msg in catalog2:
                if msg.id not in catalog1:
                    catalog1[msg.id] = msg


def create_babel_mapping(jinja_dirs=None, jinja_exts=None, extractors=None):
    exts = ",".join(jinja_exts or [])
    conf = "[python:**.py]\n"
    if jinja_dirs:
        for jinja_dir in jinja_dirs:
            if jinja_dir == '.':
                jinja_dir = ''
            conf += "[jinja2:%s]\n" % os.path.join(jinja_dir, "**.html")
            if exts:
                conf += "extensions=%s\n" % exts
    if extractors:
        for extractor, settings in extractors:
            conf += "[%s]\n" % extractor
            for k, v in settings.items():
                conf += "%s = %s\n" % (k, v)
    return conf


def exec_babel_extract(path, potfile, mapping=None, keywords=None):
    state = get_extension_state('frasco_babel')
    if mapping:
        mapping_file = tempfile.NamedTemporaryFile()
        mapping_file.write(mapping.encode('utf-8'))
        mapping_file.flush()

    if isinstance(keywords, str):
        keywords = list(map(str.strip, str(keywords).split(";")))
    elif not keywords:
        keywords = []
    keywords.extend(["_n:1,2", "translatable", "translate", "ntranslate:1,2",
                        "lazy_translate", "lazy_gettext"])
    keywords.extend(state.options['extract_keywords'])

    cmdline = [state.options["babel_bin"], "extract", "-o", potfile]
    if mapping:
        cmdline.extend(["-F", mapping_file.name])
    for k in keywords:
        cmdline.append("-k")
        cmdline.append(k)
    cmdline.append(path)

    shell_exec(cmdline)
    if mapping:
        mapping_file.close()


def po_to_json(filename):
    json_dct = {}
    with edit_pofile(filename) as catalog:
        for message in catalog:
            if not message.id:
                continue
            if message.pluralizable:
                json_dct[_format_json_msg_id(message)] = [message.id[1]] + list(message.string)
            else:
                json_dct[_format_json_msg_id(message)] = [None, message.string]
    return json.dumps(json_dct)


def _format_json_msg_id(message):
    mid = message.id[0] if message.pluralizable else message.id
    if message.context:
        return "%s::%s" % (message.context, mid)
    return mid



from wtforms import FileField as BaseFileField, ValidationError
from flask_wtf.file import FileRequired
from werkzeug.datastructures import FileStorage
from frasco.ext import get_extension_state
import os
import uuid

from .utils import generate_filename, file_size


class FileField(BaseFileField):
    def __init__(self, label=None, validators=None, auto_save=True, upload_dir=None, upload_backend=None,\
                 uuid_prefix=None, keep_filename=None, subfolders=None, backend_in_filename=True,
                 save_original_filename=None, save_file_size=None, save_mimetype=None, **kwargs):
        super(FileField, self).__init__(label, validators, **kwargs)
        self.file = None
        self.auto_save = auto_save
        self.upload_dir = upload_dir
        self._upload_backend = upload_backend
        self.uuid_prefix = uuid_prefix
        self.keep_filename = keep_filename
        self.subfolders = subfolders
        self.backend_in_filename = backend_in_filename
        self.save_original_filename = save_original_filename
        self.save_file_size = save_file_size
        self.save_mimetype = save_mimetype

    @property
    def upload_backend(self):
        return get_extension_state('frasco_upload').get_backend(self._upload_backend)

    def process_formdata(self, valuelist):
        if not valuelist:
            return
        self.file = valuelist[0]
        self.data = None
        self.filename = None
        if not self.has_file():
            return

        self.filename = generate_filename(self.file.filename,
            self.uuid_prefix, self.keep_filename, self.subfolders)
        if self.upload_dir:
            self.filename = os.path.join(self.upload_dir, self.filename)
        if self.backend_in_filename:
            self.data = self._upload_backend + '://' + self.filename
        else:
            self.data = self.filename

        if self.auto_save:
            self.save_file()

    @property
    def file_size(self):
        return file_size(self.file)

    def save_file(self):
        self.upload_backend.save(self.file, self.filename)

    def has_file(self):
        # compatibility with Flask-WTF
        if not isinstance(self.file, FileStorage):
            return False
        return self.file.filename not in [None, '', '<fdopen>']

    def populate_obj(self, obj, name):
        setattr(obj, name, self.data)
        if self.save_original_filename:
            setattr(obj, self.save_original_filename, self.file.filename)
        if self.save_file_size:
            setattr(obj, self.save_file_size, self.file_size)
        if self.save_mimetype:
            setattr(obj, self.save_mimetype, self.file.mimetype)


class FileAllowed(object):
    def __init__(self, extensions, message=None):
        self.extensions = extensions
        self.message = message

    def __call__(self, form, field):
        if not field.has_file():
            return

        filename = field.file.filename.lower()
        ext = filename.rsplit('.', 1)[-1]
        if ext in self.extensions:
            return

        message = self.message
        if message is None:
            message = field.gettext("File type not allowed")
        raise ValidationError(message)



from flask import after_this_request, _request_ctx_stack
from flask_sqlalchemy import SignallingSession
from frasco.ctx import ContextStack, DelayedCallsContext
from frasco.utils import AttrDict
from contextlib import contextmanager
from sqlalchemy import event
import functools
from .ext import db
import logging


__all__ = ('transaction', 'as_transaction', 'current_transaction', 'is_transaction', 'delayed_tx_calls', 'after_transaction_commit')


_transaction_ctx = ContextStack(default_item=True)
delayed_tx_calls = DelayedCallsContext()
logger = logging.getLogger('frasco.models')


def ensure_no_partially_rollbacked_transaction():
    if not db.session.is_active:
        db.session.rollback()


@contextmanager
def transaction(ensure_no_partially_rollbacked_tx=False, force_begin=False, nested=False):
    if ensure_no_partially_rollbacked_tx:
        ensure_no_partially_rollbacked_transaction()
    is_subtransaction = _transaction_ctx.top and not nested
    if not is_subtransaction:
        logger.debug('BEGIN TRANSACTION')
    _transaction_ctx.push()
    delayed_tx_calls.push()
    if not is_subtransaction and (force_begin or nested):
        db.session.begin(nested=nested)
    try:
        yield
        _transaction_ctx.pop()
        if not is_subtransaction:
            logger.debug('COMMIT TRANSACTION')
            db.session.commit()
        else:
            db.session.flush()
        delayed_tx_calls.pop()
    except:
        _transaction_ctx.pop()
        if not is_subtransaction:
            logger.debug('ROLLBACK TRANSACTION')
            db.session.rollback()
        delayed_tx_calls.pop(drop_calls=True)
        raise


def is_transaction():
    return bool(_transaction_ctx.top)


def as_transaction(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with transaction():
            return func(*args, **kwargs)
    return wrapper


def after_transaction_commit(func):
    delayed_tx_calls.call(func, [], {})


_current_transaction_ctx = ContextStack()
current_transaction = _current_transaction_ctx.make_proxy()


@event.listens_for(SignallingSession, 'after_transaction_create')
def on_after_transaction_create(session, transaction):
    if transaction.parent is None:
        _current_transaction_ctx.push(AttrDict())


@event.listens_for(SignallingSession, 'after_transaction_end')
def on_after_transaction_end(session, transaction):
    if transaction.parent is None:
        _current_transaction_ctx.pop()
